/*! \file eqpotgraph.cpp
 *  \brief %Graph for plotting equipotential lines
 */

/* Copyright (c) 2005-2011 Taneli Kalvas. All rights reserved.
 *
 * You can redistribute this software and/or modify it under the terms
 * of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this library (file "COPYING" included in the package);
 * if not, write to the Free Software Foundation, Inc., 51 Franklin
 * Street, Fifth Floor, Boston, MA 02110-1301 USA
 * 
 * If you have questions about your rights to use or distribute this
 * software, please contact Berkeley Lab's Technology Transfer
 * Department at TTD@lbl.gov. Other questions, comments and bug
 * reports should be sent directly to the author via email at
 * taneli.kalvas@jyu.fi.
 * 
 * NOTICE. This software was developed under partial funding from the
 * U.S.  Department of Energy.  As such, the U.S. Government has been
 * granted for itself and others acting on its behalf a paid-up,
 * nonexclusive, irrevocable, worldwide license in the Software to
 * reproduce, prepare derivative works, and perform publicly and
 * display publicly.  Beginning five (5) years after the date
 * permission to assert copyright is obtained from the U.S. Department
 * of Energy, and subject to any subsequent five (5) year renewals,
 * the U.S. Government is granted for itself and others acting on its
 * behalf a paid-up, nonexclusive, irrevocable, worldwide license in
 * the Software to reproduce, prepare derivative works, distribute
 * copies to the public, perform publicly and display publicly, and to
 * permit others to do so.
 */

#include "eqpotgraph.hpp"
#include "lineclip.hpp"
#include "ibsimu.hpp"


EqPotGraph::EqPotGraph( const ScalarField &epot, const Geometry &g )
    : _color(Color(0.2,1,0.2)), _epot(epot), _g(g), _data_built(false), _cache(true)
{
}


EqPotGraph::~EqPotGraph()
{
    for( size_t a = 0; a < _lines.size(); a++ )
	delete _lines[a];
}


void EqPotGraph::set_eqlines_manual( const std::vector<double> &pot )
{
    _data_built = false;
    _eqlines_manual = pot;
}


void EqPotGraph::set_eqlines_auto( size_t N )
{
    _data_built = false;
    _eqlines_auto = N;
}


bool EqPotGraph::eqline_exists( double pot1, signed char sol1, 
				double pot2, signed char sol2, 
				double pot ) const 
{
    if( // Basic rule: at least one point in vacuum/neumann/dirichlet 
	// and potential pot is between pot1 and pot2.
	( ((sol1 <= 6 && sol1 >= -6) || (sol2 <= 6 && sol2 >= -6)) &&
	  pot1 != pot2 &&
	  ((pot1 >= pot && pot2 <= pot) ||
	   (pot1 <= pot && pot2 >= pot)) )
	&&
	// Additional rule: potential pot has to be a real (non-virtual) potential close to 
	// solid boundaries with virtual potentials.
	!(( sol1 < -6 && 
	    ((pot1 >= pot && _g.get_boundary(-sol1).val <= pot) ||
	     (pot1 <= pot && _g.get_boundary(-sol1).val >= pot)) ) 
	  ||
	  ( sol2 < -6 && 
	    ((pot2 >= pot && _g.get_boundary(-sol2).val <= pot) ||
	     (pot2 <= pot && _g.get_boundary(-sol2).val >= pot)) )) )
    {
	return( true );
    }
    
    return( false );
}


void EqPotGraph::build_data( void )
{
    _data_built = true;

    // Clear old data
    for( size_t a = 0; a < _lines.size(); a++ )
	delete _lines[a];
    _lines.clear();

    // Build vector of required manual potential lines
    for( size_t a = 0; a < _eqlines_manual.size(); a++ )
	_lines.push_back( new EqPotLines( _eqlines_manual[a] ) );

    // Add automatic lines
    double min, max;
    _epot.epot_get_minmax( _g, min, max );
    for( size_t a = 0; a < _eqlines_auto; a++ )
	_lines.push_back( new EqPotLines( min + (a+0.5)*(max-min)/(_eqlines_auto) ) );

    // Go through mesh
    size_t i[3];
    size_t dx, dy;
    i[_vb[2]] = (int)floor(_level+0.5);
    if( _vb[0] == 0 ) dx = 1;
    else if( _vb[0] == 1 ) dx = _g.size(0);
    else dx = _g.size(1)*_g.size(0);
    if( _vb[1] == 0 ) dy = 1;
    else if( _vb[1] == 1 ) dy = _g.size(0);
    else dy = _g.size(1)*_g.size(0);
    size_t sizex = _g.size(_vb[0])-1;
    size_t sizey = _g.size(_vb[1])-1;
    for( i[_vb[1]] = 0; i[_vb[1]] < sizey; i[_vb[1]]++ ) {
	for( i[_vb[0]] = 0; i[_vb[0]] < sizex; i[_vb[0]]++ ) {
	    
	    //std::cout << "Scanning at " << i[_vb[0]] << " " << i[_vb[1]] << "\n";
	    size_t ptr = _g.size(0)*_g.size(1)*i[2] + _g.size(0)*i[1] + i[0];

	    // Potentials and solid mesh numbers counterclockwise
	    double pot1       = _epot( ptr );
	    signed char sol1  = _g.mesh( ptr );
	    size_t ptrt       = ptr+dx;
	    double pot2       = _epot( ptrt );
	    signed char sol2  = _g.mesh( ptrt );
	    ptrt             += dy;
	    double pot3       = _epot( ptrt );
	    signed char sol3  = _g.mesh( ptrt );
	    ptrt             -= dx;
	    double pot4       = _epot( ptrt );
	    signed char sol4  = _g.mesh( ptrt );

	    // Check for existance of each equipotential line
	    for( size_t a = 0; a < _lines.size(); a++ ) {

		size_t b = 0;
		double pot = _lines[a]->pot;
		double x[8];

		//std::cout << "pot = " << pot << "\n";

		// Bottom
		if( eqline_exists( pot1, sol1, pot2, sol2, pot ) ) {
		    if( pot2 != pot1 )
			x[b+0] = (i[_vb[0]]+(pot-pot1)/(pot2-pot1)) * _g.h() + _g.origo(_vb[0]);
		    else
			x[b+0] = i[_vb[0]]*_g.h() + _g.origo(_vb[0]);
		    x[b+1] = i[_vb[1]]*_g.h() + _g.origo(_vb[1]);
		    b += 2;
		}

		// Right
		if( eqline_exists( pot2, sol2, pot3, sol3, pot ) ) {
		    x[b+0] = (i[_vb[0]]+1)*_g.h() + _g.origo(_vb[0]);
		    if( pot3 != pot2 )
			x[b+1] = (i[_vb[1]]+(pot-pot2)/(pot3-pot2)) * _g.h() + _g.origo(_vb[1]);
		    else
			x[b+1] = i[_vb[1]]*_g.h() + _g.origo(_vb[1]);
		    b += 2;
		}
		
		// Top
		if( eqline_exists( pot4, sol4, pot3, sol3, pot ) ) {
		    if( pot4 != pot3 )
			x[b+0] = (i[_vb[0]]+(pot-pot4)/(pot3-pot4)) * _g.h() + _g.origo(_vb[0]);
		    else
			x[b+0] = (i[_vb[0]]+1)*_g.h() + _g.origo(_vb[0]);
		    x[b+1] = (i[_vb[1]]+1) * _g.h() + _g.origo(_vb[1]);
		    b += 2;
		}

		// Left
		if( eqline_exists( pot1, sol1, pot4, sol4, pot ) ) {
		    x[b+0] = i[_vb[0]]*_g.h() + _g.origo(_vb[0]);
		    if( pot4 != pot1 )
			x[b+1] = (i[_vb[1]]+(pot-pot1)/(pot4-pot1)) * _g.h() + _g.origo(_vb[1]);
		    else
			x[b+1] = (i[_vb[1]]+1)*_g.h() + _g.origo(_vb[1]);
		    b += 2;
		}
		
		if( b == 4 ) {
		    _lines[a]->x.push_back( Line( x[0], x[1], x[2], x[3] ) );
		} else if( b == 8 ) {
		    // Six possible connections possible for four
		    // points We don't want to draw across. 
		    // Equipotential lines don't cross -> out
		    // of four lines we draw two
		    _lines[a]->x.push_back( Line( x[0], x[1], x[2], x[3] ) );
		    _lines[a]->x.push_back( Line( x[4], x[5], x[6], x[7] ) );
		}
	    }
	}
    }
}


void EqPotGraph::disable_cache( void )
{
    _cache = false;
}


void EqPotGraph::plot( cairo_t *cairo, const Coordmapper *cm, const double range[4] )
{
    if( !_data_built || _oview != _view || _olevel != _level || !_cache ) {
	// First round or change happened
	//_ilevel = (int)floor( (_level - _g.origo(_vb[2])) / _g.h() + 0.5 );
	build_data();
    }
    _oview = _view;
    _olevel = _level;

    // Set drawing properties
    cairo_set_source_rgba( cairo, _color[0], _color[1], _color[2], _color[3] );
    cairo_set_line_width( cairo, 1.0 );
    LineClip lc( cairo );

    for( size_t a = 0; a < _lines.size(); a++ ) {
	for( size_t b = 0; b < _lines[a]->x.size(); b++ ) {

	    double xout[2];
	    cm->transform( xout, &_lines[a]->x[b][0] );
	    //std::cout << "From:"  << xout[0] << " " << xout[1] << "\n";
	    lc.move_to( xout[0], xout[1] );
	    cm->transform( xout, &_lines[a]->x[b][2] );
	    //std::cout << "To:  " << xout[0] << " " << xout[1] << "\n";
	    lc.line_to( xout[0], xout[1] );
	    cairo_stroke( cairo );
	}
    }
}


void EqPotGraph::plot_sample( cairo_t *cairo, double x, double y, double width, double height )
{

}


void EqPotGraph::get_bbox( double bbox[4] )
{
    bbox[0] = _g.origo( _vb[0] );
    bbox[1] = _g.origo( _vb[1] );
    bbox[2] = _g.max( _vb[0] );
    bbox[3] = _g.max( _vb[1] );
}




















